from netbox.api.routers import NetBoxRouter

from netbox_more_metrics.api import views

router = NetBoxRouter()
router.register("collections", views.MetricCollectionViewSet)
router.register(
    "metric-value-type-options",
    views.MetricValueTypeOptionsViewSet,
    basename="metric_value_type_options",
)
urlpatterns = router.urls



from django import forms
from django.shortcuts import reverse
from utilities.forms.widgets import APISelect

from netbox_more_metrics.choices import MetricValueChoices


class DynamicMetricValueOptionField(forms.ChoiceField):
    widget = APISelect

    def __init__(self, query_params=None, object_type_field=None, *args, **kwargs):
        self.query_params = query_params or {}
        self.object_type_field = object_type_field

        # These are always valid and should be the default.
        kwargs["choices"] = MetricValueChoices.DEFAULT_CHOICES
        kwargs["initial"] = MetricValueChoices.DEFAULT_CHOICE

        super().__init__(*args, **kwargs)

    def get_bound_field(self, form, field_name):
        bound_field = forms.BoundField(form, self, field_name)
        widget = bound_field.field.widget

        # Attach any query parameters
        if len(self.query_params) > 0:
            widget.add_query_params(self.query_params)

        widget.attrs["data-url"] = reverse(
            "plugins-api:netbox_more_metrics-api:metric_value_type_options-list"
        )

        return bound_field



from prometheus_client import Counter

__all__ = ("dynamic_metric_collectors",)

dynamic_metric_collectors = Counter(
    "dynamic_metric_collectors",
    "Information about active dynamic metric collectors.",
    ["id", "metric_name"],
)



from netbox.api.serializers import NetBoxModelSerializer, WritableNestedSerializer
from rest_framework import serializers

from netbox_more_metrics.models import Metric, MetricCollection


class MetricCollectionSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netbox_more_metrics-api:metriccollection-detail"
    )

    class Meta:
        model = MetricCollection
        fields = ("id", "name", "display", "enabled", "url")


class NestedMetricCollectionSerializer(WritableNestedSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netbox_more_metrics-api:metriccollection-detail"
    )

    class Meta:
        model = MetricCollection
        fields = ("id", "display", "url")


class MetricSerializer(NetBoxModelSerializer):
    class Meta:
        model = Metric
        fields = ("id", "name")


class MetricValueOptionSerializer(serializers.Serializer):
    id = (
        serializers.CharField()
    )  # We 'abuse' the id field here so that the normal APISelect widget/js works.
    display = serializers.CharField()



from django.core.validators import RegexValidator, ValidationError
from django.utils.regex_helper import _lazy_re_compile
from django.utils.translation import gettext_lazy as _

metric_name_re = _lazy_re_compile(r"^[a-zA-Z_][a-zA-Z0-9_]+\Z")
validate_metric_name = RegexValidator(
    metric_name_re,
    _("Enter a valid “metric_name” consisting of letters, numbers or underscores."),
    "invalid",
)

label_name_re = _lazy_re_compile(r"^[a-zA-Z_][a-zA-Z0-9_]+\Z")
validate_label_name = RegexValidator(
    label_name_re,
    _("Enter a valid “label_name” consisting of letters, numbers or underscores."),
    "invalid",
)


def validate_label_renames(value):
    """
    Validator to ensure that the JSONField only contains key-value pairs where both keys and values are strings.
    Nested structures are not allowed.
    """
    if not isinstance(value, dict):
        raise ValidationError(
            _("Invalid data type: %(value)s. Expected a dictionary."),
            params={"value": value},
        )

    seen_label_names = set()
    for key, val in value.items():
        if not isinstance(key, str):
            raise ValidationError(
                _("Invalid key type: %(key)s. Expected a string."),
                params={"key": key},
            )
        if not isinstance(val, str):
            raise ValidationError(
                _("Invalid value type for key '%(key)s': %(val)s. Expected a string."),
                params={"key": key, "val": val},
            )

        if val in seen_label_names:
            raise ValidationError(
                _("Duplicate label name: %(val)s."),
                params={"val": val},
            )

        # Add the label name to the set of seen label names
        seen_label_names.add(val)

        # Check if the value is a valid label name
        validate_label_name(val)



from django.db import migrations, models

import netbox_more_metrics.validators


class Migration(migrations.Migration):

    dependencies = [
        ("netbox_more_metrics", "0003_alter_metric_content_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="metric",
            name="label_renames",
            field=models.JSONField(
                blank=True,
                default=dict,
                validators=[netbox_more_metrics.validators.validate_label_renames],
            ),
        ),
    ]



from django.db.backends.signals import connection_created
from netbox.plugins import PluginConfig

from netbox_more_metrics.signals import start_collectors
from netbox_more_metrics.utilities import enable_metrics


class NetBoxMoreMetricsConfig(PluginConfig):
    name = "netbox_more_metrics"
    verbose_name = "More Metrics"
    description = "Export custom metrics from NetBox data."
    version = "0.3.2"
    min_version = "4.0"
    author = "Daan van Gorkum"
    author_email = "me+netbox@dj.vg"
    base_url = "more-metrics"

    def ready(self):
        # Make sure we call the NetBox plugin initialization to add the menus etc.
        super().ready()

        # Only enable the global metrics if we want to and the application is ready to serve them.
        if enable_metrics():
            connection_created.connect(start_collectors)


config = NetBoxMoreMetricsConfig



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0010_gfk_indexes"),
        ("netbox_more_metrics", "0002_metric_metric_value"),
    ]

    operations = [
        migrations.AlterField(
            model_name="metric",
            name="content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.objecttype",
            ),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("netbox_more_metrics", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="metric",
            name="metric_value",
            field=models.CharField(default="count", max_length=50),
        ),
    ]



from contextlib import suppress

from django.db.backends.signals import connection_created
from django.db.utils import ProgrammingError
from prometheus_client import REGISTRY


def start_collectors(sender, **kwargs):
    with suppress(ProgrammingError):
        from netbox_more_metrics.collectors import DynamicMetricCollectionCollector
        from netbox_more_metrics.metrics import dynamic_metric_collectors  # noqa: F401

        DynamicMetricCollectionCollector(registry=REGISTRY)

    # Disconnect the signal, so we don't run it again.
    connection_created.disconnect(start_collectors)



import logging
from time import time
from typing import Iterable

from django.core.exceptions import FieldError
from django.db.models import CharField, Count, F, Func, Value
from django.db.models.functions import Cast, Coalesce
from prometheus_client import metrics_core
from prometheus_client.core import (
    CounterMetricFamily,
    GaugeMetricFamily,
    InfoMetricFamily,
)
from prometheus_client.registry import REGISTRY, Collector, CollectorRegistry

from netbox_more_metrics.metrics import dynamic_metric_collectors
from netbox_more_metrics.models import Metric, MetricCollection

logger = logging.getLogger(__name__)


class DynamicMetricCollectionCollector(Collector):
    def __init__(
        self,
        registry: CollectorRegistry = REGISTRY,
        collection: MetricCollection = None,
    ):
        self._is_default_registry = registry is REGISTRY
        self.registry = registry

        self.base_queryset = (
            collection.metrics.all() if collection else Metric.objects.all()
        )

        filter = dict(enabled=True, collections__enabled=True)

        if self._is_default_registry:
            filter.update(dict(collections__include_in_default=True))

        self.queryset = self.base_queryset.filter(**filter)

        if self._is_default_registry:
            logger.debug("Collector %r for default registry.", self)

        # Start the collectors.
        self._refresh_collectors()

        # We only register ourselves so that we can check for metrics that need to be added.
        # The actual metrics are emitted from their own instance and unregister themselves if needed.
        self.registry.register(self)

    def _get_enabled_metrics(self):
        metrics = self.queryset.all().distinct()
        return metrics

    def _refresh_collectors(self):
        for metric in self._get_enabled_metrics():
            if metric.metric_name not in self.registry._names_to_collectors:
                logger.debug(
                    "Adding dynamic metric collector for Metric '%s' (%d).",
                    metric.metric_name,
                    metric.pk,
                )
                DynamicMetricCollector(metric=metric, registry=self.registry)

    def collect(self) -> Iterable[metrics_core.Metric]:
        logger.debug("Refreshing dynamic metric collectors...")
        self._refresh_collectors()
        return

        # We have to yield some metric otherwise this will fail in the registry .collect() method.
        yield InfoMetricFamily()


class DynamicMetricCollector(Collector):
    def __init__(
        self,
        metric: Metric,
        registry: CollectorRegistry = REGISTRY,
        force_enable: bool = False,
    ):
        self._is_default_registry = registry is REGISTRY
        self._metric = metric
        self._registry = registry
        self.force_enable = force_enable

        self.pk = metric.pk
        self.name = metric.metric_name
        self.description = metric.metric_description
        self.metric_family = self._metric.metric_family
        self.metric_value = self._metric.metric_value or "count"
        self.model = self._metric.content_type.model_class()
        self.queryset = self.model.objects.all()

        # FIXME would be nice if we can put this in the choices someway.
        self.metric_value_is_method = not self.metric_value == "count"

        self.filter = self._metric.filter
        self.labels = {
            field: self._metric.label_renames.get(field, field)
            for field in self._metric.metric_labels
        }
        self.created = time()

        self._internal_labels = (str(self.pk), self.name)

        logger.debug(
            "Dynamic metric collector for Metric '%s' (%d).",
            self._metric.metric_name,
            self._metric.pk,
        )

        # If the filter/labels are not valid we should not start this collector.
        if (self.filter and not self.test_filter()) or not self.test_labels():
            return

        registry.register(self)

        # Add an info metric of known dynamic collectors if it's attached to the global registry.
        if self._is_default_registry:
            dynamic_metric_collectors.labels(*self._internal_labels)

    def test_filter(self):
        # Test the filter to make sure it's valid.
        try:
            self.queryset.filter(**self.filter)
        except FieldError:
            logger.exception(
                "Metric '%s' (%d) filter is not valid.", self.name, self.pk
            )
            return False

        return True

    def test_labels(self):
        # Test the labels make sure they're valid.
        try:
            self.queryset.values(*self.labels.keys())
        except FieldError:
            logger.exception("Metric '%s' (%d) labels are invalid.", self.name, self.pk)
            return False

        return True

    def unregister(self):
        logger.debug(
            "Unregistering dynamic metric collector '%s' (%d).", self.name, self.pk
        )
        self._registry.unregister(self)
        dynamic_metric_collectors.remove(*self._internal_labels)

    def renew(self):
        self.unregister()
        return self.__class__(registry=self._registry, metric=self._metric)

    def get_queryset(self):
        qs = self.queryset
        if self.filter:
            qs = qs.filter(**self.filter)

        return qs

    def get_label_annotations(self):
        annotations = {}
        for field, label_name in self.labels.items():
            field_name = f"__metric_label_{label_name}"
            if field.startswith("custom_field_data__"):
                keys = [Value(key) for key in field.split("__")[1:]]
                annotations[field_name] = Coalesce(
                    Func(
                        F("custom_field_data"),
                        *keys,
                        function="jsonb_extract_path_text",
                        output_field=CharField(),
                    ),
                    Value("null"),
                )
            else:
                annotations[field_name] = Coalesce(
                    Cast(F(field), output_field=CharField()), Value("null")
                )
        return annotations

    def get_metric_result(self):
        label_annotations = self.get_label_annotations()
        values = label_annotations.keys()

        if self.metric_value_is_method:
            return self.get_queryset().annotate(count=Count("*"), **label_annotations)

        base_qs = (
            self.get_queryset()
            .values(*self.labels.keys())
            .annotate(count=Count("*"), **label_annotations)
            .values("count", *values)
        )

        return base_qs

    def is_metric_enabled(self):
        if not self._metric.enabled:
            return False

        if self._is_default_registry:
            return bool(
                self._metric.collections.filter(enabled=True, include_in_default=True)
            )

        return bool(self._metric.collections.filter(enabled=True))

    def collect(self) -> Iterable[metrics_core.Metric]:
        logger.debug("Collecting for Metric '%s' (%d).", self.name, self.pk)

        # Run some basic checks if this metric is still healthy.
        if self._is_default_registry:
            try:
                self._metric.refresh_from_db()
            except Metric.DoesNotExist:
                # If the object has disappeared we unregister ourselves.
                self.unregister()
                return

        # Check if the metric has changed, and if that's the case renew the metric.
        # Yield the metrics from the new instance.
        if self.created < self._metric.last_updated.timestamp():
            yield from self.renew().collect()
            return

        # Metric has been disabled, unregister.
        if not self.force_enable and not self.is_metric_enabled():
            if self._is_default_registry:
                self.unregister()
            return

        # Create the metric
        metric: GaugeMetricFamily | CounterMetricFamily | InfoMetricFamily = (
            self.metric_family(self.name, self.description, labels=self.labels)
        )

        # Get the data for the metric.
        results = self.get_metric_result()

        for result in results:
            if isinstance(result, dict):
                labels = {
                    field: result.get(f"__metric_label_{field}")
                    for field in self.labels.values()
                }
                value = result.pop(self.metric_value)
            else:
                labels = {
                    field: getattr(result, f"__metric_label_{field}")
                    for field in self.labels.values()
                }
                value = (
                    getattr(result, self.metric_value)
                    if not self.metric_value_is_method
                    else getattr(result, self.metric_value)()
                )

            if self.metric_family is InfoMetricFamily:
                metric.add_metric(labels="", value=labels)
            else:
                metric.add_metric(labels.values(), value)

        yield metric

        if self._is_default_registry:
            dynamic_metric_collectors.labels(*self._internal_labels).inc()

    def describe(self):
        yield self.metric_family(self.name, self.description)



import django_tables2 as tables
from netbox.tables import NetBoxTable, columns

from netbox_more_metrics.models import Metric, MetricCollection


class MetricCollectionTable(NetBoxTable):
    name = tables.Column(linkify=True)
    include_in_default = tables.BooleanColumn(verbose_name="Exported globally")

    class Meta(NetBoxTable.Meta):
        model = MetricCollection
        fields = ("id", "name", "enabled", "include_in_default", "description")
        default_columns = ("pk", "name", "enabled", "include_in_default", "description")


class MetricTable(NetBoxTable):
    name = tables.Column(linkify=True)
    metric_name = tables.Column(verbose_name="Metric")
    content_type = columns.ContentTypeColumn(verbose_name="Object type")
    metric_description = tables.Column(verbose_name="Description")
    metric_type = columns.ChoiceFieldColumn()

    class Meta(NetBoxTable.Meta):
        model = Metric
        fields = (
            "id",
            "name",
            "enabled",
            "metric_name",
            "labels",
            "metric_type",
            "content_type",
            "metric_description",
        )
        default_columns = (
            "pk",
            "name",
            "enabled",
            "metric_name",
            "labels",
            "metric_type",
            "content_type",
            "metric_description",
        )



from django.urls import path
from netbox.views.generic import ObjectChangeLogView, ObjectJournalView

from netbox_more_metrics import views
from netbox_more_metrics.models import Metric, MetricCollection

urlpatterns = [
    #
    # MetricCollection
    #
    path(
        "collections/",
        views.MetricCollectionListView.as_view(),
        name="metriccollection_list",
    ),
    path(
        "collections/add/",
        views.MetricCollectionEditView.as_view(),
        name="metriccollection_add",
    ),
    path(
        "collections/<int:pk>/",
        views.MetricCollectionView.as_view(),
        name="metriccollection",
    ),
    path(
        "collections/<int:pk>/edit/",
        views.MetricCollectionEditView.as_view(),
        name="metriccollection_edit",
    ),
    path(
        "collections/<int:pk>/metrics/",
        views.MetricCollectionExportView.as_view(),
        name="metriccollection_metrics",
    ),
    path(
        "collections/<int:pk>/delete/",
        views.MetricCollectionDeleteView.as_view(),
        name="metriccollection_delete",
    ),
    path(
        "collections/<int:pk>/changelog/",
        ObjectChangeLogView.as_view(),
        name="metriccollection_changelog",
        kwargs={"model": MetricCollection},
    ),
    path(
        "collections/<int:pk>/journal/",
        ObjectJournalView.as_view(),
        name="metriccollection_journal",
        kwargs={"model": MetricCollection},
    ),
    #
    # Metric
    #
    path("metrics/", views.MetricListView.as_view(), name="metric_list"),
    path("metrics/add/", views.MetricEditView.as_view(), name="metric_add"),
    path("metrics/<int:pk>/", views.MetricView.as_view(), name="metric"),
    path("metrics/<int:pk>/edit/", views.MetricEditView.as_view(), name="metric_edit"),
    path(
        "metrics/<int:pk>/metrics/",
        views.MetricExportView.as_view(),
        name="metric_metrics",
    ),
    path(
        "metrics/<int:pk>/delete/",
        views.MetricDeleteView.as_view(),
        name="metric_delete",
    ),
    path(
        "metrics/<int:pk>/changelog/",
        ObjectChangeLogView.as_view(),
        name="metric_changelog",
        kwargs={"model": Metric},
    ),
    path(
        "metrics/<int:pk>/journal/",
        ObjectJournalView.as_view(),
        name="metric_journal",
        kwargs={"model": Metric},
    ),
]



from django.db.models import F, Func, TextField, Value
from django.http import HttpResponse
from netbox.views.generic import (
    ObjectDeleteView,
    ObjectEditView,
    ObjectListView,
    ObjectView,
)
from prometheus_client import CONTENT_TYPE_LATEST, CollectorRegistry, generate_latest

from netbox_more_metrics.collectors import (
    DynamicMetricCollectionCollector,
    DynamicMetricCollector,
)
from netbox_more_metrics.forms import MetricCollectionForm, MetricForm
from netbox_more_metrics.models import Metric, MetricCollection
from netbox_more_metrics.tables import MetricCollectionTable, MetricTable

#
# MetricCollection
#


class MetricCollectionListView(ObjectListView):
    queryset = MetricCollection.objects.all()
    actions = {"add": {"add"}}
    table = MetricCollectionTable


class MetricCollectionView(ObjectView):
    queryset = MetricCollection.objects.all()

    def get_extra_context(self, request, instance):
        metrics_table = MetricTable(
            data=instance.metrics.restrict(request.user, "view").annotate(
                labels=Func(
                    F("metric_labels"),
                    Value(", "),
                    function="array_to_string",
                    output_field=TextField(),
                )
            ),
            orderable=False,
            exclude=("actions", "labels"),
        )

        return {"metrics_table": metrics_table}


class MetricCollectionExportView(ObjectView):
    queryset = MetricCollection.objects.all()
    http_method_names = ("get",)

    def get(self, request, **kwargs):
        instance = self.get_object(**kwargs)

        registry = CollectorRegistry()
        DynamicMetricCollectionCollector(registry=registry, collection=instance)

        metrics_page = generate_latest(registry)

        return HttpResponse(metrics_page, content_type=CONTENT_TYPE_LATEST)


class MetricCollectionEditView(ObjectEditView):
    queryset = MetricCollection.objects.all()
    form = MetricCollectionForm


class MetricCollectionDeleteView(ObjectDeleteView):
    queryset = MetricCollection.objects.all()


#
# Metric
#


class MetricListView(ObjectListView):
    queryset = Metric.objects.all().annotate(
        labels=Func(
            F("metric_labels"),
            Value(", "),
            function="array_to_string",
            output_field=TextField(),
        )
    )
    actions = {"add": {"add"}}
    table = MetricTable


class MetricView(ObjectView):
    queryset = Metric.objects.all()

    def get_extra_context(self, request, instance):
        # Generate metrics to be shown on page.
        registry = CollectorRegistry()
        DynamicMetricCollector(registry=registry, metric=instance, force_enable=True)

        # Collections table
        collections_table = MetricCollectionTable(
            data=instance.collections.restrict(request.user, "view"),
            orderable=False,
            exclude=("actions",),
        )

        return {
            "metrics": generate_latest(registry).decode(),
            "collections_table": collections_table,
        }


class MetricExportView(ObjectView):
    queryset = Metric.objects.all()
    http_method_names = ("get",)

    def get(self, request, **kwargs):
        instance = self.get_object(**kwargs)

        registry = CollectorRegistry()
        DynamicMetricCollector(registry=registry, metric=instance)

        metrics_page = generate_latest(registry)

        return HttpResponse(metrics_page, content_type=CONTENT_TYPE_LATEST)


class MetricEditView(ObjectEditView):
    queryset = Metric.objects.all()
    form = MetricForm


class MetricDeleteView(ObjectDeleteView):
    queryset = Metric.objects.all()



from netbox.api.viewsets import NetBoxModelViewSet
from rest_framework import viewsets
from rest_framework.response import Response

from netbox_more_metrics.api.serializers import (
    MetricCollectionSerializer,
    MetricValueOptionSerializer,
)
from netbox_more_metrics.choices import MetricValueChoices
from netbox_more_metrics.models import MetricCollection


class MetricCollectionViewSet(NetBoxModelViewSet):
    queryset = MetricCollection.objects.all()
    serializer_class = MetricCollectionSerializer


class MetricValueTypeOptionsViewSet(viewsets.ViewSet):
    """
    Return valid Metric Value Type for an Object Type.
    """

    permission_classes = []

    def list(self, request):
        content_type = request.query_params.get("object_type")
        choices = MetricValueChoices.DEFAULT_CHOICES
        if content_type:
            choices = MetricValueChoices.choices_for_contenttype(content_type)

        options = list()
        for option in choices:
            value, name = option
            options.append({"id": value, "display": name})

        serializer = MetricValueOptionSerializer(options, many=True)

        return Response(
            {
                "count": len(options),
                "next": None,
                "previous": None,
                "results": serializer.data,
            }
        )



from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import FieldError, ValidationError
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext as _
from netbox.models import NetBoxModel

from netbox_more_metrics.choices import MetricTypeChoices
from netbox_more_metrics.validators import (
    validate_label_name,
    validate_label_renames,
    validate_metric_name,
)


class ObjectAbsoluteUrlMixin:
    def get_absolute_url(self):
        path = f"plugins:{self._meta.app_label}:{self._meta.model_name}"
        return reverse(path, args=[self.pk])


class MetricCollection(ObjectAbsoluteUrlMixin, NetBoxModel):
    """
    Model that represents a CollectorRegistry.
    You can connect Metric instances to this to export only these specific metrics.
    """

    name = models.CharField(max_length=50)
    description = models.CharField(max_length=255, blank=True)
    enabled = models.BooleanField(default=True)
    include_in_default = models.BooleanField(
        default=False, help_text=_("Include collection in default metric REGISTRY.")
    )

    def __str__(self):
        return self.name


class Metric(ObjectAbsoluteUrlMixin, NetBoxModel):
    """
    Represents a single Metric to be exported.
    """

    name = models.CharField(max_length=50)
    enabled = models.BooleanField(default=True)
    metric_name = models.CharField(
        unique=True, max_length=50, validators=[validate_metric_name]
    )
    metric_description = models.CharField(max_length=255)
    metric_labels = ArrayField(
        base_field=models.CharField(max_length=50, validators=[validate_label_name])
    )
    metric_type = models.CharField(max_length=50, choices=MetricTypeChoices)
    metric_value = models.CharField(max_length=50, default="count")
    content_type = models.ForeignKey(
        to="core.ObjectType",
        related_name="+",
        verbose_name="Object type",
        help_text=_("The object to which this Metric applies."),
        on_delete=models.CASCADE,
    )
    filter = models.JSONField(
        null=False, default=dict, blank=True, help_text=_("QuerySet filter")
    )
    label_renames = models.JSONField(
        null=False,
        default=dict,
        blank=True,
        help_text=_("Label renaming"),
        validators=[validate_label_renames],
    )
    collections = models.ManyToManyField(to=MetricCollection, related_name="metrics")

    def __str__(self):
        return self.name

    def clean(self):
        super().clean()

        # Sort the labels, so they're in the same order as the exported metrics.
        self.metric_labels.sort()

        model = self.content_type.model_class()

        # Test the labels
        try:
            model.objects.values(*self.metric_labels)
        except FieldError as e:
            raise ValidationError({"metric_labels": f"Labels invalid: {e}"})

        # Test the filter
        if self.filter:
            try:
                model.objects.filter(**self.filter)
            except FieldError as e:
                raise ValidationError({"filter": f"Filter invalid: {e}"})
        else:
            self.filter = {}

        if self.label_renames:
            # Make sure all renamed labels exists as labels
            for label in self.label_renames:
                if label not in self.metric_labels:
                    raise ValidationError(
                        {
                            "label_renames": f"Label '{label}' is not in the metric labels."
                        }
                    )

            # Make sure there's no duplicate label names
            for label in self.metric_labels:
                # First check if the label is in the label_renames:
                if label in self.label_renames.values():
                    # If it is it's possible that the label is being renamed to something else, if not, trigger a
                    # ValidationError
                    if (
                        label not in self.label_renames
                        or label == self.label_renames.get(label)
                    ):
                        raise ValidationError(
                            {
                                "label_renames": f"Label '{label}' is being shadowed by a label rename."
                            }
                        )

    @property
    def metric_family(self):
        return MetricTypeChoices.TYPES[self.metric_type]



import re
import sys

from django.conf import settings


def enable_metrics():
    if hasattr(settings, "METRICS_ENABLED") and not settings.METRICS_ENABLED:
        return False

    # Enable metrics on the devserver
    if "runserver" in sys.argv:
        return True

    # Enable metrics if not running manage.py.
    regex = re.compile(r"(?:\.\/)?manage\.py", flags=re.IGNORECASE)
    if not any(filter(regex.match, sys.argv)):
        return True

    return False



from core.models import ObjectType
from netbox.choices import ChoiceSet
from prometheus_client.metrics_core import (
    CounterMetricFamily,
    GaugeMetricFamily,
    InfoMetricFamily,
)


class MetricTypeChoices(ChoiceSet):
    TYPE_GAUGE = "gauge"
    TYPE_COUNTER = "counter"
    TYPE_INFO = "info"

    CHOICES = ((TYPE_COUNTER, "Counter"), (TYPE_INFO, "Info"), (TYPE_GAUGE, "Gauge"))

    TYPES = {
        TYPE_INFO: InfoMetricFamily,
        TYPE_COUNTER: CounterMetricFamily,
        TYPE_GAUGE: GaugeMetricFamily,
    }


class MetricValueChoices(ChoiceSet):
    DEFAULT_CHOICES = (("count", "Count"),)
    DEFAULT_CHOICE = "count"

    CHOICES = (
        (
            "dcim.Rack",
            (
                ("get_utilization", "Percentage Utilized"),
                ("get_power_utilization", "Power Percentage Utilized"),
            ),
        ),
        ("ipam.Prefix", (("get_utilization", "Percentage Utilized"),)),
        (
            "ipam.Aggregate",
            (("get_utilization", "Percentage Utilized"),),
        ),
    )
    CHOICES_BY_MODEL = dict(CHOICES)

    @classmethod
    def choices_for_model(cls, app_label: str, model_label: str) -> tuple:
        choices = cls.CHOICES_BY_MODEL.get(f"{app_label}.{model_label}", ())
        return tuple(choices) + cls.DEFAULT_CHOICES

    @classmethod
    def choices_for_contenttype(cls, model) -> tuple:
        if isinstance(model, (str, int)):
            try:
                model = ObjectType.objects.get_for_id(model).model_class()
            except ObjectType.DoesNotExist:
                return cls.DEFAULT_CHOICES

        app_name, model_name = model._meta.label.split(".")
        return cls.choices_for_model(app_name, model_name)



from netbox.plugins import PluginMenuButton, PluginMenuItem

metriccollection_buttons = [
    PluginMenuButton(
        link="plugins:netbox_more_metrics:metriccollection_add",
        title="Add",
        icon_class="mdi mdi-plus-thick",
        permissions=("netbox_more_metrics.add_metriccollection",),
    )
]

metric_buttons = [
    PluginMenuButton(
        link="plugins:netbox_more_metrics:metric_add",
        title="Add",
        icon_class="mdi mdi-plus-thick",
        permissions=("netbox_more_metrics.add_metric",),
    )
]

menu_items = (
    PluginMenuItem(
        link="plugins:netbox_more_metrics:metriccollection_list",
        link_text="Metric Collections",
        buttons=metriccollection_buttons,
    ),
    PluginMenuItem(
        link="plugins:netbox_more_metrics:metric_list",
        link_text="Metrics",
        buttons=metric_buttons,
    ),
)



from core.models import ObjectType
from django import forms
from netbox.forms import NetBoxModelForm
from utilities.forms.fields import (
    ContentTypeChoiceField,
    DynamicModelMultipleChoiceField,
    JSONField,
)
from utilities.forms.rendering import FieldSet

from netbox_more_metrics.choices import MetricValueChoices
from netbox_more_metrics.fields import DynamicMetricValueOptionField
from netbox_more_metrics.models import Metric, MetricCollection


class MetricCollectionForm(NetBoxModelForm):
    include_in_default = forms.BooleanField(label="Exported globally", required=False)

    class Meta:
        model = MetricCollection
        fields = ("name", "description", "enabled", "include_in_default", "tags")


class MetricForm(NetBoxModelForm):
    collections = DynamicModelMultipleChoiceField(
        queryset=MetricCollection.objects.all()
    )
    content_type = ContentTypeChoiceField(
        label="Object Type", queryset=ObjectType.objects.all()
    )
    metric_description = forms.CharField(label="Description")

    metric_value = DynamicMetricValueOptionField(
        query_params={"object_type": "$content_type"},
        object_type_field="content_type",
        help_text="Select the value used for the metric. This might ignore aggregation done by labels.",
    )

    label_renames = JSONField(
        label="Label renaming",
        help_text="Rename label names to make them clearer. "
        "For example, 'device__manufacturer_slug' to 'manufacturer':"
        """<pre>{
    "device__manufacturer_slug": "manufacturer"
}</pre>""",
        required=False,
    )

    fieldsets = (
        FieldSet("name", "metric_description", "enabled", "tags"),
        FieldSet("content_type", "filter", name="Metric source"),
        FieldSet(
            "metric_name",
            "metric_labels",
            "label_renames",
            "metric_type",
            "metric_value",
            name="Metric configuration",
        ),
        FieldSet("collections", name="Metric exposition"),
    )

    class Meta:
        model = Metric
        fields = (
            "name",
            "metric_description",
            "enabled",
            "metric_name",
            "metric_labels",
            "metric_type",
            "metric_value",
            "label_renames",
            "filter",
            "content_type",
            "collections",
        )
        field_classes = {
            "filter": JSONField,
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set the choices for the content_type field.
        if self.data:
            content_type = self.data.get("content_type")
            if content_type:
                self.fields["metric_value"].choices = (
                    MetricValueChoices.choices_for_contenttype(content_type)
                )
        elif self.instance.pk:
            content_type = self.instance.content_type.model_class()
            self.fields["metric_value"].choices = (
                MetricValueChoices.choices_for_contenttype(content_type)
            )



import re

import django.contrib.postgres.fields
import django.core.validators
import django.db.models.deletion
import taggit.managers
import utilities.json
from django.db import migrations, models

import netbox_more_metrics.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("extras", "0084_staging"),
    ]

    operations = [
        migrations.CreateModel(
            name="MetricCollection",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("name", models.CharField(max_length=50)),
                ("description", models.CharField(blank=True, max_length=255)),
                ("enabled", models.BooleanField(default=True)),
                ("include_in_default", models.BooleanField(default=False)),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model, netbox_more_metrics.models.ObjectAbsoluteUrlMixin),
        ),
        migrations.CreateModel(
            name="Metric",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("name", models.CharField(max_length=50)),
                ("enabled", models.BooleanField(default=True)),
                (
                    "metric_name",
                    models.CharField(
                        max_length=50,
                        unique=True,
                        validators=[
                            django.core.validators.RegexValidator(
                                re.compile("^[a-zA-Z_][a-zA-Z0-9_]+\\Z"),
                                "Enter a valid “metric_name” consisting of letters, numbers or underscores.",
                                "invalid",
                            )
                        ],
                    ),
                ),
                ("metric_description", models.CharField(max_length=255)),
                (
                    "metric_labels",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=50,
                            validators=[
                                django.core.validators.RegexValidator(
                                    re.compile("^[a-zA-Z_][a-zA-Z0-9_]+\\Z"),
                                    "Enter a valid “label_name” consisting of letters, numbers or underscores.",
                                    "invalid",
                                )
                            ],
                        ),
                        size=None,
                    ),
                ),
                ("metric_type", models.CharField(max_length=50)),
                ("filter", models.JSONField(blank=True, default=dict)),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="metrics",
                        to="netbox_more_metrics.metriccollection",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model, netbox_more_metrics.models.ObjectAbsoluteUrlMixin),
        ),
    ]
